// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package main

import (
	"flag"
	"fmt"
	"os"
	"os/signal"
	"path/filepath"

	"golang.org/x/sys/unix"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/internal/manager"
	"sr.ht/~craftyguy/superd/internal/misc"
	"sr.ht/~craftyguy/superd/internal/server"
)

// set at build time
var Version string

func usage() {
	flag.CommandLine.Usage()
}

func main() {
	// this depends on XDG_RUNTIME_DIR, so make sure it is set and fail fast if
	// not
	if v, ok := os.LookupEnv("XDG_RUNTIME_DIR"); v == "" || !ok {
		panic("$XDG_RUNTIME_DIR is not set in the environment, this is required.")
	} else {
		// Only allow running 1 instance of superd for the current user
		lockFile := filepath.Join(v, "superd.lock")
		f, err := os.Create(lockFile)
		if err != nil {
			fmt.Printf("unable to create lock file at %q: %s\n", lockFile, err)
			os.Exit(1)
		}

		if err := unix.Flock(int(f.Fd()), unix.LOCK_EX|unix.LOCK_NB); err == unix.EWOULDBLOCK {
			fmt.Println("unable to acquire lock file, is another instance of superd running?")
			os.Exit(1)
		}

		defer func() {
			unix.Flock(int(f.Fd()), unix.LOCK_UN)
			f.Close()
			os.RemoveAll(lockFile)
		}()
	}

	var serviceDir string
	flag.StringVar(&serviceDir, "d", "", "Specify an additional directory to scan for service config")
	var logFilePath string
	flag.StringVar(&logFilePath, "l", "", "File location to log output to (default: \"$XDG_STATE_HOME/superd.log\")")
	var logVerbose bool
	flag.BoolVar(&logVerbose, "v", false, "Enable verbose debug output to log")
	var showVersion bool
	flag.BoolVar(&showVersion, "version", false, "Print version and quit.")
	var showHelp bool
	flag.BoolVar(&showHelp, "h", false, "Print help and quit.")

	flag.Usage = func() {
		fmt.Println("usage: superd [OPTION...]")
		fmt.Println("Options:")
		flag.PrintDefaults()
	}

	flag.Parse()

	if showHelp {
		usage()
		return
	}

	if showVersion {
		fmt.Printf("%s - %s\n", filepath.Base(os.Args[0]), Version)
		os.Exit(0)
	}

	// setup log file
	if logFilePath == "" {
		logDir, ok := os.LookupEnv("XDG_STATE_HOME")
		if !ok {
			logDir = "~/.local/state"
		}
		logFilePath = filepath.Join(logDir, "superd.log")
	}
	logFilePath = misc.ExpandHome(logFilePath)
	logDir := filepath.Dir(logFilePath)

	if err := os.MkdirAll(logDir, 0750); err != nil {
		panic(fmt.Sprintf("unable to create dir %q for log file: %s", logDir, err))
	}

	// rotate any old log file to <name>.old
	// note: error intentionally ignored, if we can't write to the dir that the
	// log file is in, we'll find out soon enough anyways when creating the new
	// log file
	_ = os.Rename(logFilePath, logFilePath+".old")

	logFile, err := os.OpenFile(misc.ExpandHome(logFilePath), os.O_CREATE|os.O_RDWR|os.O_TRUNC, 0644)
	if err != nil {
		panic(err)
	}
	defer logFile.Close()

	logger := log.New(logFile)
	if logVerbose {
		logger.SetMinLevel(log.DEBUG)
	}

	mgr := manager.Init(serviceDir)

	stopChan := make(chan struct{})

	// start signal handler
	sigChan := make(chan os.Signal, 1)
	signal.Notify(sigChan, unix.SIGTERM, unix.SIGQUIT, unix.SIGINT)

	go func() {
		// block until some signal is recieved
		<-sigChan

		close(stopChan)
	}()

	// start api server on unix socket
	// TODO: make path configurable?
	// TODO: add option to disable starting this server?
	socketPath := "$XDG_RUNTIME_DIR/superd.sock"
	server := server.New(os.ExpandEnv(socketPath))
	go func() {
		if err := server.Start(mgr); err != nil {
			panic(err)
		}
	}()

	// Main "loop'
	if err := mgr.Start(stopChan); err != nil {
		panic(err)
	}
}
