// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package dag

import (
	"errors"
)

type DAG struct {
	nodes map[string][]string
}

var ErrDAGCycle = errors.New("cycle detected")

func NewDag() (d *DAG) {
	// nodes stored in adjacency list
	return &DAG{
		nodes: make(map[string][]string),
	}
}

func (d *DAG) AddEdge(fromNode string, toNode string) error {
	// if node is already in adjacency list for fromNode, skip it
	for _, n := range d.nodes[fromNode] {
		if n == toNode {
			return nil
		}
	}

	// don't add an edge if it creates a cycle
	if err := d.checkCycle(fromNode, toNode); err != nil {
		return err
	}

	d.nodes[fromNode] = append(d.nodes[fromNode], toNode)
	return nil
}

// get the list of adjacent nodes to the given node
func (d *DAG) AdjacentNodes(node string) (list []string) {
	if _, ok := d.nodes[node]; !ok {
		return
	}
	return d.nodes[node]
}

func (d *DAG) TopoSort() (sorted []string) {
	visited := make(map[string]bool)
	for n := range d.nodes {
		visited[n] = false
	}

	var revSorted []string

	for n := range d.nodes {
		if !visited[n] {
			d.topoSort(n, &visited, &revSorted)
		}
	}

	for i := len(revSorted) - 1; i >= 0; i-- {
		sorted = append(sorted, revSorted[i])
	}

	return
}

func (d *DAG) topoSort(node string, visited *map[string]bool, sorted *[]string) {
	(*visited)[node] = true

	for _, n := range d.nodes[node] {
		if !(*visited)[n] {
			d.topoSort(n, visited, sorted)
		}
	}

	(*sorted) = append(*sorted, node)
}

// Checks if adding an edge between the two given nodes will create a cycle in
// the DAG. Returns ErrDAGCycle if a cycle would be created.
func (d *DAG) checkCycle(fromNode string, toNode string) (err error) {
	visited := make(map[string]bool)

	for n := range d.nodes {
		visited[n] = false
	}

	if err := d.checkCycleDFS(toNode, &visited, fromNode); err != nil {
		return err
	}

	return
}

func (d *DAG) checkCycleDFS(node string, visited *map[string]bool, ancestor string) error {
	(*visited)[node] = true

	for _, n := range d.nodes[node] {
		if ancestor == n {
			return ErrDAGCycle
		} else if !(*visited)[n] {
			if err := d.checkCycleDFS(n, visited, ancestor); err != nil {
				return err
			}
		}
	}

	return nil
}
