// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package dag

import (
	"errors"
	"reflect"
	"sort"
	"testing"
)

func helperBuildDag(in [][2]string) *DAG {
	d := NewDag()

	for _, e := range in {
		d.AddEdge(e[0], e[1])
	}

	return d
}

func TestTopoSort(t *testing.T) {
	tables := []struct {
		name string
		// slice of slices, which are edges
		in [][2]string
		// topo sorted list of nodes, as a list of possible lists since there
		// are multiple solutions for ordering for some DAGs
		expected [][]string
	}{
		{
			name: "simple, 1 solution",
			in: [][2]string{
				{"A", "B"},
				{"C", "B"},
			},
			expected: [][]string{
				{"A", "C", "B"},
				{"C", "A", "B"},
			},
		},
		{
			name: "simple, 2 solution",
			in: [][2]string{
				{"A", "B"},
				{"A", "C"},
				{"B", "D"},
				{"C", "D"},
			},
			expected: [][]string{
				{"A", "B", "C", "D"},
				{"A", "C", "B", "D"},
			},
		},
		{
			name: "many solutions",
			in: [][2]string{
				{"A", "B"},
				{"C", "B"},
				{"B", "D"},
				{"C", "E"},
				{"D", "F"},
				{"E", "F"},
			},
			expected: [][]string{
				{"A", "C", "B", "D", "E", "F"},
				{"A", "C", "B", "E", "D", "F"},
				{"A", "C", "E", "B", "D", "F"},
				{"C", "A", "B", "D", "E", "F"},
				{"C", "A", "B", "E", "D", "F"},
				{"C", "A", "E", "B", "D", "F"},
				{"C", "E", "A", "B", "D", "F"},
			},
		},
	}

	for _, table := range tables {

		t.Run(table.name, func(t *testing.T) {
			d := helperBuildDag(table.in)

			r := d.TopoSort()

			pass := false
			for _, e := range table.expected {
				if reflect.DeepEqual(r, e) {
					pass = true
				}
			}
			if !pass {
				t.Fatalf("Expected: %q, got: %q", table.expected, r)
			}
		})
	}
}

func TestCheckCycle(t *testing.T) {
	tables := []struct {
		name string
		// slice of slices, which are edges
		in      [][2]string
		in_edge [2]string
		// true if cycle, else false
		expected bool
	}{
		{
			name: "simple no cycle",
			in: [][2]string{
				{"A", "B"},
				{"B", "C"},
				{"A", "C"},
			},
			in_edge:  [2]string{"A", "C"},
			expected: false,
		},
		{
			name: "simple direct cycle",
			in: [][2]string{
				{"A", "B"},
			},
			in_edge:  [2]string{"B", "A"},
			expected: true,
		},
		{
			name: "large indirect cycle",
			in: [][2]string{
				{"A", "B"},
				{"C", "B"},
				{"B", "D"},
				{"C", "A"},
				{"C", "E"},
				{"D", "F"},
				{"E", "F"},
			},
			in_edge:  [2]string{"F", "B"},
			expected: true,
		},
		{
			name: "simple indirect cycle",
			in: [][2]string{
				{"A", "B"},
				{"A", "C"},
				{"D", "A"},
			},
			in_edge:  [2]string{"B", "D"},
			expected: true,
		},
	}

	for _, table := range tables {
		t.Run(table.name, func(t *testing.T) {
			d := helperBuildDag(table.in)

			hasCycle := false
			if err := d.checkCycle(table.in_edge[0], table.in_edge[1]); err != nil {
				if errors.Is(err, ErrDAGCycle) {
					hasCycle = true
				} else {
					t.Fatalf("Unexpected error: %s", err)
				}
			}

			if hasCycle != table.expected {
				t.Fatalf("Expected cycle: %t, got: %t", table.expected, hasCycle)
			}
		})
	}
}

func TestAdjacentNodes(t *testing.T) {
	tables := []struct {
		name string
		// slice of slices, which are edges
		in      [][2]string
		in_node string
		// true if cycle, else false
		expected []string
	}{
		{
			name: "0 adjacent",
			in: [][2]string{
				{"A", "B"},
				{"B", "C"},
				{"A", "C"},
			},
			in_node: "Z",
		},
		{
			name: "2 adjacent",
			in: [][2]string{
				{"A", "B"},
				{"B", "C"},
				{"A", "C"},
			},
			in_node:  "A",
			expected: []string{"B", "C"},
		},
		{
			name: "6 adjacent",
			in: [][2]string{
				{"A", "F"},
				{"A", "D"},
				{"A", "C"},
				{"A", "E"},
				{"A", "B"},
				{"A", "G"},
				{"B", "G"},
				{"Y", "Z"},
			},
			in_node:  "A",
			expected: []string{"B", "C", "D", "E", "F", "G"},
		},
		{
			name: "no duplicates",
			in: [][2]string{
				{"A", "B"},
				{"A", "B"},
				{"A", "C"},
			},
			in_node:  "A",
			expected: []string{"B", "C"},
		},
	}

	for _, table := range tables {
		t.Run(table.name, func(t *testing.T) {
			d := helperBuildDag(table.in)

			// sorted to make comparison straightforward, and there's no
			// guarantee of order for adjacent nodes in a DAG
			out := d.AdjacentNodes(table.in_node)
			sort.Strings(out)
			sort.Strings(table.expected)

			if !reflect.DeepEqual(out, table.expected) {
				t.Fatalf("Expected %v, got: %v", table.expected, out)
			}
		})
	}

}
