// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package log

import (
	"fmt"
	"io"
	stdLog "log"
	"os"

	"github.com/hashicorp/logutils"
)

type Logger struct {
	log    *stdLog.Logger
	filter *logutils.LevelFilter
}

const (
	DEBUG logutils.LogLevel = "DEBUG"
	ERROR logutils.LogLevel = "ERROR"
	INFO  logutils.LogLevel = "INFO"
	WARN  logutils.LogLevel = "WARN"
)

var logger *Logger

// Initialize a new logger writing the given io.Writer. This is a singleton, so
// if New() was called previously, the previously initialized object is
// returned.
func New(out io.Writer) *Logger {
	if logger == nil {
		logger = &Logger{
			filter: &logutils.LevelFilter{
				Levels:   []logutils.LogLevel{DEBUG, INFO, WARN, ERROR},
				MinLevel: logutils.LogLevel(INFO),
				Writer:   out,
			},
		}
		logger.log = stdLog.New(logger.filter, "", stdLog.Ldate|stdLog.Ltime)
	}
	return logger
}

// Initialize a new logger writing to stderr. This is a singleton, so if New()
// was called previously, then that object will be returned and not necessarily
// one initialized to write to stderr.
func Default() *Logger {
	return New(os.Stderr)
}

func Get() *Logger {
	return logger
}

// Sets the minimum logging level
func (l *Logger) SetMinLevel(level logutils.LogLevel) {
	l.filter.SetMinLevel(level)
}

func (l *Logger) Printf(p string, f string, s ...interface{}) {
	l.log.Print(format(p, INFO, fmt.Sprintf(f, s...)))
}

func (l *Logger) Println(p string, s ...interface{}) {
	l.log.Println(format(p, INFO, s...))
}

func (l *Logger) Debugf(p string, f string, s ...interface{}) {
	l.log.Print(format(p, DEBUG, fmt.Sprintf(f, s...)))
}

func (l *Logger) Debugln(p string, s ...interface{}) {
	l.log.Println(format(p, DEBUG, s...))
}

func (l *Logger) Warnf(p string, f string, s ...interface{}) {
	l.log.Print(format(p, WARN, fmt.Sprintf(f, s...)))
}

func (l *Logger) Warnln(p string, s ...interface{}) {
	l.log.Println(format(p, WARN, s...))
}

func (l *Logger) Errorf(p string, f string, s ...interface{}) {
	l.log.Print(format(p, ERROR, fmt.Sprintf(f, s...)))
}

func (l *Logger) Errorln(p string, s ...interface{}) {
	l.log.Println(format(p, ERROR, s...))
}

func format(p string, l logutils.LogLevel, s ...interface{}) string {
	return fmt.Sprintf("%s [%s]: ", p, l) + fmt.Sprint(s...)
}
