// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package server

import (
	"encoding/json"
	"fmt"
	"net"
	"os"
	"strings"
	"time"

	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/internal/manager"
	"sr.ht/~craftyguy/superd/pkg/api"
)

var logger *log.Logger
var prefix = "server"

type Server struct {
	socketPath string
	socket     net.Listener
}

// Create a new Server listening on the given unix socket
func New(socketPath string) (s *Server) {
	logger = log.Get()

	s = &Server{
		socketPath: socketPath,
	}

	return
}

func (s *Server) Start(mgr *manager.Manager) (err error) {
	if err := os.RemoveAll(s.socketPath); err != nil {
		return fmt.Errorf("server.Start failed: %w", err)
	}

	s.socket, err = net.Listen("unix", s.socketPath)
	if err != nil {
		return fmt.Errorf("server.Start failed: %w", err)
	}
	defer s.socket.Close()

	if err := os.Chmod(s.socketPath, 0600); err != nil {
		return fmt.Errorf("server.Start failed: %w", err)
	}

	if err := os.Chown(s.socketPath, os.Getuid(), os.Getgid()); err != nil {
		return fmt.Errorf("server.Start failed: %w", err)
	}

	for {
		conn, err := (s.socket).Accept()
		if err != nil {
			return fmt.Errorf("server.Start failed: %w", err)
		}

		go s.clientConnection(&conn, mgr)
	}
}

func (s *Server) clientConnection(c *net.Conn, mgr *manager.Manager) {
	defer (*c).Close()

	// prevent bad clients from "hanging" the goroutine
	// TODO: this has to be reset every blocking read/write...
	// TODO: also it might be too aggressive?
	(*c).SetDeadline(time.Now().Add(2 * time.Second))

	reqBuf := make([]byte, 256)
	size, err := (*c).Read(reqBuf)
	if err != nil {
		logger.Errorln(prefix, err.Error())
		return
	}

	var req api.Request
	// Note: reqBuf is re-sliced to be equal in length to number of bytes read,
	// or unmarshalling fails
	if err := json.Unmarshal(reqBuf[:size], &req); err != nil {
		logger.Errorf(prefix, fmt.Sprintf("unable to unmarshal request: %s", string(reqBuf)))
		logger.Errorf(prefix, fmt.Sprintf("unmarshal error: %s", err))
		return
	}

	var resp api.Response

	switch req.Cmd {
	case api.CmdStatus:
		status, err := mgr.Details(req.Params)
		if err != nil {
			resp = api.ServiceDetailsResponse{
				ErrorResponse: api.NewErrorResponse("Status: " + err.Error()),
			}
			break
		}

		resp = api.ServiceDetailsResponse{
			Result: status,
		}

	case api.CmdStart:
		// 'start' only expects 1 parameter, the service name
		if len(req.Params) != 1 {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Start service: unexpected number of params specified, expected only 1, received %d", len(req.Params))),
			}
			break
		}
		err := mgr.StartService(req.Params[0])
		if err != nil {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.NilResponse{}

	case api.CmdStop:
		// 'stop' only expects 1 parameter, the service name
		if len(req.Params) != 1 {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Stop service: unexpected number of params specified, expected only 1, received %d", len(req.Params))),
			}
			break
		}
		err := mgr.StopService(req.Params[0])
		if err != nil {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.NilResponse{}

	case api.CmdRestart:
		// 'restart' only expects 1 parameter, the service name
		if len(req.Params) != 1 {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Restart service: unexpected number of params specified, expected only 1, received %d", len(req.Params))),
			}
			break
		}

		err := mgr.RestartService(req.Params[0])
		if err != nil {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.NilResponse{}

	case api.CmdEnable:
		// 'enable' only expects 1 parameter, the service name
		if len(req.Params) != 1 {
			resp = api.StringResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Enable service: unexpected number of params specified, expected only 1, received %d", len(req.Params))),
			}
			break
		}

		result, err := mgr.EnableService(req.Params[0])
		if err != nil {
			resp = api.StringResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.StringResponse{
			Result: result,
		}

	case api.CmdDisable:
		// 'disable' only expects 1 parameter, the service name
		if len(req.Params) != 1 {
			resp = api.StringResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Disable service: unexpected number of params specified, expected only 1, received %d", len(req.Params))),
			}
			break
		}

		result, err := mgr.DisableService(req.Params[0])
		if err != nil {
			resp = api.StringResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.StringResponse{
			Result: result,
		}

	case api.CmdSetEnv:
		// 'set-env' expects a list
		if len(req.Params) < 1 {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse("set environment: expects at least 1 parameter"),
			}
			break
		}

		env := make(map[string]string)

		for _, val := range req.Params {
			parts := strings.Split(val, "=")
			if len(parts) != 2 {
				continue
			}

			env[parts[0]] = parts[1]
		}

		mgr.SetEnv(env)

		resp = api.NilResponse{}

	case api.CmdReload:
		// 'reload' expects 0 args
		if len(req.Params) != 0 {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(fmt.Sprintf("Reload: no parameters expected, received: %d", len(req.Params))),
			}
			break
		}

		err := mgr.Reload()
		if err != nil {
			resp = api.NilResponse{
				ErrorResponse: api.NewErrorResponse(err.Error()),
			}
			break
		}

		resp = api.NilResponse{}

	default:
		resp = api.NilResponse{
			ErrorResponse: api.NewErrorResponse("unknown command received: " + string(req.Cmd)),
		}
		logger.Errorf(prefix, "unknown command received: "+string(req.Cmd))
	}

	j, err := resp.Json()
	logger.Debugf(prefix, string(j))
	if err != nil {
		logger.Errorf(prefix, "unable to marshal response: "+err.Error())
	}

	(*c).SetDeadline(time.Now().Add(2 * time.Second))
	_, err = (*c).Write(j)
	if err != nil {
		logger.Errorf(prefix, "unable to send response: "+err.Error())
	}
}
