// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"os"
	"time"

	goCmd "github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/env"
)

type procDetails struct {
	proc       *goCmd.Cmd
	procStatus <-chan goCmd.Status
	stdout     <-chan string
	stderr     <-chan string
}

// Runs the given exe w/ args, returns the status and the details/properties of
// the started process if it was successful.
func (s *Service) runCmd(exe string, args []string) (cmd *procDetails, ok bool) {

	var (
		startTimer <-chan time.Time
		c          *goCmd.Cmd
	)

	cmd = &procDetails{}
	c = execCommand(goCmd.Options{
		Streaming: true,
		Buffered:  false,
	}, exe)
	cmd.proc = c
	c.Args = args

	// set service-specific env. from config
	for e, v := range s.config.Env {
		c.Env = append(c.Env, fmt.Sprintf("%s=%s", e, v))
	}

	// add "global" environment
	e := env.Init()
	for v := range e.IterVars() {
		c.Env = append(c.Env, v)
	}

	// import all variables from current env
	c.Env = append(c.Env, os.Environ()...)

	cmd.procStatus = c.Start()
	cmd.stderr = c.Stderr
	cmd.stdout = c.Stderr

	startTimer = time.After(startTimerDuration)

	logger.Println(s.Name(), "running exe: ", exe)

	select {
	// Start timer
	case <-startTimer:
		// start timer expired, if exe is still up then it's probably fine
		startTimer = nil

		if s.isUp(c.Status().PID) {
			ok = true
		} else {
			logger.Warnln(s.Name(), "timed out waiting for command to start: ", exe, args)
			ok = false
		}

	case pStatus := <-cmd.procStatus:
		// status from go-cmd
		if pStatus.Complete {
			logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(pStatus.Exit))
			if pStatus.Exit != 0 {
				ok = false
			} else if s.config.Type == config.TypeOneshot {
				// show 'started' for oneshot services
				ok = true
			}
		} else {
			logger.Warnln(s.Name(), "received unhandled status: "+fmt.Sprint(s))
			ok = false
		}
	}

	return
}
