// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"os"
	"sync"
	"syscall"
	"time"

	goCmd "github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Service struct {
	Start   chan bool
	Stop    chan bool
	Details chan api.ServiceDetails
	Config  chan *config.ServiceConfig

	before struct {
		list []string
		sync.RWMutex
	}
	after struct {
		list []string
		sync.RWMutex
	}

	name struct {
		value string
		sync.RWMutex
	}

	autostart struct {
		value bool
		sync.RWMutex
	}

	config *config.ServiceConfig
}

// whether or not the service should be autostarted
func (s *Service) Autostart() bool {
	s.autostart.RLock()
	defer s.autostart.RUnlock()
	return s.autostart.value
}

func (s *Service) Name() string {
	s.name.RLock()
	defer s.name.RUnlock()
	return s.name.value
}

func (s *Service) AddAfter(name string) {
	s.after.Lock()
	defer s.after.Unlock()
	for _, n := range s.after.list {
		if n == name {
			return
		}
	}
	s.after.list = append(s.after.list, name)
}

func (s *Service) AddBefore(name string) {
	s.before.Lock()
	defer s.before.Unlock()
	for _, n := range s.before.list {
		if n == name {
			return
		}
	}
	s.before.list = append(s.before.list, name)
}

func (s *Service) After() []string {
	s.after.RLock()
	defer s.after.RUnlock()
	// TODO: is this really a deep copy?
	return s.after.list
}

func (s *Service) Before() []string {
	s.before.RLock()
	defer s.before.RUnlock()
	return s.before.list
}

var logger *log.Logger

func Init(name string) (s *Service) {
	logger = log.Get()

	s = &Service{
		Start:   make(chan bool),
		Stop:    make(chan bool),
		Details: make(chan api.ServiceDetails),
		Config:  make(chan *config.ServiceConfig),
	}
	s.name.Lock()
	defer s.name.Unlock()
	s.name.value = name

	return
}

var (
	// for mocking in tests
	execCommand = goCmd.NewCmdOptions
	// give 1 second to startup
	// TODO: use systemd.service TimeoutSec, default to 1(?)
	startTimerDuration = 1 * time.Second
)

// startup service routine, e.g. running ExecStartPre, ExecStart. Returns
// api.ServiceFailed on failure, or a serviceProc with references to the
// running process if startup was successful.
func (s *Service) startup() (sProc chan *procDetails, startStatus chan api.ServiceStatus) {
	startStatus = make(chan api.ServiceStatus)
	sProc = make(chan *procDetails)

	go func() {
		defer close(startStatus)
		defer close(sProc)

		// ExecStartPre
		if s.config.ExecStartPreCmd != "" {
			_, ok := s.runCmd(s.config.ExecStartPreCmd, s.config.ExecStartPreCmdArgs)
			if !ok {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// ExecStart
		cmd, ok := s.runCmd(s.config.Cmd, s.config.CmdArgs)
		if !ok {
			startStatus <- api.ServiceFailed
			return
		}

		// ExecStartPost
		if s.config.ExecStartPostCmd != "" {
			_, ok := s.runCmd(s.config.ExecStartPostCmd, s.config.ExecStartPostCmdArgs)
			if !ok {
				startStatus <- api.ServiceFailed
				return
			}
		}

		// send the cmd from ExecStart
		sProc <- cmd
	}()

	return
}

func (s *Service) Run(quit <-chan struct{}) {

	// TODO: should be a manager config option
	// maxRetries := 10

	var (
		stdout         <-chan string
		stderr         <-chan string
		procStatus     <-chan goCmd.Status
		startStatus    <-chan api.ServiceStatus
		onFailComplete chan struct{}
		restartTimer   <-chan time.Time
		details        api.ServiceDetails
		// tries      int

		// used for receiving the started service process details
		procChan chan *procDetails
		p        *procDetails
	)
	details.Status = api.ServiceStopped

	for {
		select {

		// Start (process) requested
		case <-s.Start:

			if details.Status == api.ServiceStarted || details.Status == api.ServiceStarting {
				continue
			}

			procChan, startStatus = s.startup()
			details.Status = api.ServiceStarting

		// new service process info received (e.g. from startup())
		case p = <-procChan:
			startStatus = nil
			procChan = nil
			procStatus = p.procStatus
			logger.Errorf("procStatus: ", fmt.Sprintf("%v", procStatus))
			details.Status = api.ServiceStarted
			details.Pid = p.proc.Status().PID
			stderr = p.stderr
			stdout = p.stdout

		// service startup status
		case ss := <-startStatus:
			startStatus = nil
			procChan = nil
			procStatus = nil
			switch ss {
			case api.ServiceFailed:
				details.Status = api.ServiceFailed
				// Run execOnFailure
				if s.config.ExecOnFailureCmd != "" {
					// Note: a restart is not scheduled until after the
					// ExecOnFailure command has completed
					onFailComplete = make(chan struct{})
					s.execOnFail(onFailComplete)
				} else {
					restartTimer = s.scheduleRestart(details.Status)
				}
			}

		// Stop (process) requested
		case <-s.Stop:
			// set channel to nil to avoid triggering a restart
			procStatus = nil
			details.Status = api.ServiceStopped

			// guard against many rapid calls to stop
			if p == nil {
				continue
			}

			if err := p.proc.Stop(); err != nil {
				logger.Errorln(s.Name(), "stop() returned error: "+err.Error())
				details.Status = api.ServiceFailed
			}
			p = nil

			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			if restartTimer != nil {
				restartTimer = nil
			}

		// Quit (goroutine) requested
		case <-quit:
			// set channel to nil to avoid triggering a restart
			procStatus = nil
			if p != nil {
				if err := p.proc.Stop(); err != nil {
					logger.Errorln(s.Name(), "failed to stop: "+err.Error())
				}
			}
			return

		// Send details/status
		case s.Details <- details:
			continue

		// Status from running process
		case sStatus := <-procStatus:
			// status from go-cmd
			if sStatus.Complete {
				logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(sStatus.Exit))
				if sStatus.Exit != 0 {
					details.Status = api.ServiceFailed

					// fire off any execOnFailure
					if s.config.ExecOnFailureCmd != "" {
						onFailComplete = make(chan struct{})
						s.execOnFail(onFailComplete)

						// don't schedule a restart until after the
						// ExecOnFailure command has completed
						continue
					}
				} else if s.config.Type == config.TypeOneshot {
					// show 'started' for oneshot services
					details.Status = api.ServiceStarted
				} else {
					details.Status = api.ServiceStopped
				}

				// Restart scheduled if necessary
				restartTimer = s.scheduleRestart(details.Status)
			} else {
				logger.Warnln(s.Name(), "received unhandled status: "+fmt.Sprint(s))
			}

		// onFailStatus -- from ExecOnFailure
		case <-onFailComplete:
			onFailComplete = nil
			// Restart scheduled if necessary
			restartTimer = s.scheduleRestart(details.Status)

		// Restart timer
		case <-restartTimer:
			go func() {
				// TODO: ewwwww
				s.Start <- true
			}()

		// Log output from command
		case line, open := <-stderr:
			if !open {
				stderr = nil
				continue
			}
			logger.Debugln(s.Name(), line)
		case line, open := <-stdout:
			if !open {
				stdout = nil
				continue
			}
			logger.Debugln(s.Name(), line)

		// New config received
		case c := <-s.Config:
			s.config = c

			// update details. note that this is only updating fields that
			// aren't changed by starting/stopping the service, or other
			// runtime events related to the supervised service.
			details.Autostart = c.Autostart
			details.ConfigPath = c.Path
			details.Type = string(c.Type)

			// update the autostart field, which is thread safe
			s.autostart.Lock()
			s.autostart.value = c.Autostart
			s.autostart.Unlock()
		}
	}
}

func (s *Service) isUp(pid int) bool {
	up := true

	p, err := os.FindProcess(pid)
	if err != nil {
		up = false
	} else {
		if err := p.Signal(syscall.Signal(0)); err != nil {
			up = false
		}
	}

	return up
}

func (s *Service) oneshotShouldRestart(status api.ServiceStatus) bool {
	// oneshot should only restart if it is set to Always or OnFailure, and
	// only when the status is "failed"
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	}
	return false
}

func (s *Service) simpleShouldRestart(status api.ServiceStatus) bool {
	if status == api.ServiceFailed {
		if s.config.Restart == config.RestartAlways || s.config.Restart == config.RestartFailure {
			return true
		}
	} else if status == api.ServiceStopped {
		if s.config.Restart == config.RestartAlways {
			return true
		}
	}

	return false
}

// Checks whether a restart should be scheduled based on the given status and
// the service's config, and returns a channel to a timer for when it should be
// restarted. The "timer" returned will have been started.
func (s *Service) scheduleRestart(status api.ServiceStatus) (timer <-chan time.Time) {
	var restart bool

	switch s.config.Type {
	case config.TypeOneshot:
		restart = s.oneshotShouldRestart(status)
	case config.TypeSimple:
		restart = s.simpleShouldRestart(status)
	default:
		logger.Warnf(s.Name(), "unable to determine restart behavior, unknown service type: %s", s.config.Type)
	}

	if restart {
		logger.Debugln(s.Name(), fmt.Sprintf("restarting in %d second(s)", uint64(s.config.RestartDur.Seconds())))
		timer = time.After(s.config.RestartDur)
		return
	}
	return
}

// Run the ExecOnFailure cmd and close the given channel when done
func (s *Service) execOnFail(done chan struct{}) {
	s.runCmd(s.config.ExecOnFailureCmd, s.config.ExecOnFailureCmdArgs)
	close(done)
}
