// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"os"
	"strconv"
	"testing"
	"time"

	"github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

var mockExitCode = 0

func helperService(name string) *Service {
	s := Init(name)
	s.config = &config.ServiceConfig{
		Name: name,
		Type: config.TypeSimple,
		Env: map[string]string{
			"GO_WANT_HELPER_PROCESS": "1",
			"EXIT_CODE":              strconv.Itoa(mockExitCode),
		},
		Restart: config.RestartNever,
		Cmd:     "foo",
		CmdArgs: []string{"-test.run=TestHelperExec"},
	}

	return s
}

func testExec(options cmd.Options, name string, args ...string) *cmd.Cmd {
	cmd := cmd.NewCmdOptions(options, os.Args[0])
	cmd.Env = []string{
		"GO_WANT_HELPER_PROCESS=1",
		"EXIT_CODE=" + strconv.Itoa(mockExitCode),
	}
	return cmd
}

func TestServiceStatus(t *testing.T) {
	subtests := []struct {
		name     string
		inRet    int
		inName   string
		expected api.ServiceStatus
	}{
		{
			name:     "status starting",
			inName:   "testStatusStarting",
			inRet:    0,
			expected: api.ServiceStarting,
		},
		{
			name:     "status started",
			inName:   "testStatusStarted",
			inRet:    0,
			expected: api.ServiceStarted,
		},
		{
			name:     "status failed",
			inName:   "testStatusFailed",
			inRet:    1,
			expected: api.ServiceFailed,
		},
		{
			name:     "status stopped",
			inName:   "testStatusStopped",
			inRet:    0,
			expected: api.ServiceStopped,
		},
	}

	execCommand = testExec
	startTimerDuration = 100 * time.Millisecond
	defer func() {
		execCommand = cmd.NewCmdOptions
		startTimerDuration = 1 * time.Second
	}()

	logger = log.Default()

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			quit := make(chan struct{})

			mockExitCode = subtest.inRet
			s := helperService(subtest.inName)
			go s.Run(quit)

			select {
			case s.Start <- true:
			case <-time.After(1 * time.Second):
				t.Fatalf("timed out trying to start test process")
			}

			start := time.Now()

		wait_loop:
			for d := range s.Details {
				if d.Status == subtest.expected {
					break wait_loop
				}

				// can't use time.After here in the select because reading from
				// s.Status is a fire hose and continuously reading from it will
				// spin the loop too quickly and receiving from the time.After chan
				// will won't happen (or won't happen for a loooong time)
				if time.Since(start) > 2*time.Second {
					t.Fatalf("timed out waiting for expected status: %s", subtest.expected)
				}
				time.Sleep(time.Millisecond * 1)
			}
			close(quit)
		})
	}
}

func TestHelperExec(t *testing.T) {
	if os.Getenv("GO_WANT_HELPER_PROCESS") != "1" {
		return
	}

	time.Sleep(500 * time.Millisecond)
	ret, _ := strconv.Atoi(os.Getenv("EXIT_CODE"))
	os.Exit(ret)
}

func TestShouldRestart(t *testing.T) {
	subtests := []struct {
		name      string
		inStatus  api.ServiceStatus
		inType    config.Type
		inRestart config.RestartCond
		expected  bool
	}{
		{
			name:      "simple failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple failed never restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartNever,
			expected:  false,
		},
		{
			name:      "simple failed restart on failure",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartFailure,
			expected:  true,
		},
		{
			name:      "simple starting always restart",
			inStatus:  api.ServiceStarting,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "oneshot started always restart",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot started restart on failure",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartFailure,
			expected:  false,
		},
	}

	logger = log.Default()

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			s := helperService(st.name)
			s.config.Type = st.inType
			s.config.Restart = st.inRestart

			var out bool
			switch st.inType {
			case config.TypeSimple:
				out = s.simpleShouldRestart(st.inStatus)
			case config.TypeOneshot:
				out = s.oneshotShouldRestart(st.inStatus)
			default:
				t.Fatalf("unexpected type %s", st.inType)
			}
			if out != st.expected {
				t.Fatalf("expected: %t, received: %t", st.expected, out)
			}
		})
	}
}
