// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package api

import "encoding/json"

type ServiceStatus string

const (
	ServiceFailed   ServiceStatus = "failed"
	ServiceStarted  ServiceStatus = "started"
	ServiceStarting ServiceStatus = "starting"
	ServiceStopped  ServiceStatus = "stopped"
)

type ServiceDetails struct {
	Status ServiceStatus
	Pid    int

	// path to the service's config file
	ConfigPath string

	// whether or not the service is autostarted
	Autostart bool

	Type string
}

type ReqCommand string

const (
	CmdDisable ReqCommand = "disable"
	CmdEnable  ReqCommand = "enable"
	CmdReload  ReqCommand = "reload"
	CmdRestart ReqCommand = "restart"
	CmdSetEnv  ReqCommand = "set-env"
	CmdStart   ReqCommand = "start"
	CmdStatus  ReqCommand = "status"
	CmdStop    ReqCommand = "stop"
)

type Request struct {
	Cmd    ReqCommand `json:"cmd"`
	Params []string   `json:"params"`
}

func (r Request) Json() (req []byte, err error) {
	return json.Marshal(r)
}

type Response interface {
	Json() ([]byte, error)
}

type ErrorResponse struct {
	Error    bool
	ErrorMsg string
}

func NewErrorResponse(message string) ErrorResponse {
	return ErrorResponse{
		Error:    true,
		ErrorMsg: message,
	}
}

type NilResponse struct {
	ErrorResponse `json:"error"`
}

func (r NilResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}

type ServiceDetailsResponse struct {
	Result        map[string]ServiceDetails `json:"result"`
	ErrorResponse `json:"error"`
}

func (r ServiceDetailsResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}

type StringResponse struct {
	Result        string `json:"result"`
	ErrorResponse `json:"error"`
}

func (r StringResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}
